
#include "inc.h"

#include "lib.h"
#include "printf.h"
#include "iwram_arm.h"
#include "div_arm.h"

#include "main.h"

//BG FontgꍇARgAEg폜
#include "header\font\BGFont.h"


//Mode0
//BG2 \
//BG3 ͕
//Spr J[\
//---------------------------------------------------------------------------
#define STATE_INIT 				1
#define STATE_RESET				2
#define STATE_INPUT				3
#define STATE_EDIT				4
#define STATE_RUN				5

#define BRAIN_MAX_MEM			256

//---------------------------------------------------------------------------
typedef struct {
	u16 x;
	u16 y;
	u16 cx;
	u16 cy;
	s16 minX;
	s16 maxX;
	s16 minY;
	s16 maxY;
	u16 sprNo;
	u16 sprTile;
	u16 sprSize;
	u16 sprForm;
} ST_WRITER;

typedef struct {
	u8  x;
	u8  y;
	u8  cx;
	u8  cy;
	s16 minX;
	s16 maxX;
//	s16 minY;
//	s16 maxY;
	u16 sprNo;
	u16 sprTile;
	u16 sprSize;
	u16 sprForm;
} ST_CURSOR;

typedef struct {
	u8 txt[BG_SCREEN_CY][BG_SCREEN_CX];		//SRAMҏWpu8ɕύXB{u16iBGChrTCYjł
} ST_SCREEN;

typedef struct {
	u8  point;
	u8  mem[BRAIN_MAX_MEM];
	s16 deepCnt;
} ST_BRAIN;

//---------------------------------------------------------------------------
ST_BG     BG2;
ST_BG     BG3;
ST_CURSOR Cursor;
ST_WRITER Writer;
ST_SCREEN Screen;
ST_BRAIN  Brain;

u8  State;
u16 Key;
//---------------------------------------------------------------------------
EWRAM_CODE int main()
{
	State = STATE_INIT;

	for(;;)
	{
		switch(State)
		{
			case STATE_INIT:  StateInit();  break;
			case STATE_RESET: StateReset(); break;
			case STATE_INPUT: StateInput(); break;
			case STATE_RUN:   StateRun();   break;

			default:
				DebugMessage("Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void StateInit()
{
	InitWS();
	InitKey();
	InitBG();
	InitSprite();
	InitIRQ();

	SetMode(MODE_0 | BG2_ENABLE | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);

	State = STATE_RESET;
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitWS()
{
	REG_WSCNT = 0x4317;
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitKey()
{
	KeyInit();
	Key = 0x0000;
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitBG()
{
	//BG2 ͕
	//BG3 \

	BG2.chr = 1;
	BG2.map = 0;
	BG2.col = 0;

	BG3.chr = 1;	//BG2Ɠ
	BG3.map = 1;
	BG3.col = 0;	//BG2Ɠ

	REG_BG2CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG2.chr) | BG_MAPBASE(BG2.map) | 0);

	REG_BG3CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG3.chr) | BG_MAPBASE(BG3.map) | 0);

	BGSetPalData16((u16*)&BGFont_palette, BG3.col);
	BGSetChrData  ((u16*)&BGFont_data, BGFont_DATA_SIZE, BG3.chr);
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitSprite()
{
	SpriteInit();

	SpriteSetPalData16((u16*)&sprPal);
	SpriteSetData ((u16*)&sprTiles, SPR_IMG_BYTE);
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitIRQ()
{
	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;

	REG_TM0D     = 65536 - 2621;			// (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;

	REG_IME      = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
EWRAM_CODE void StateReset()
{
	ScreenInit();
	CursorInit();
	WriterInit();

	CursorDraw();
	WriterDraw();

	State = STATE_INPUT;
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenInit()
{
	ST_SCREEN* p = &Screen;

	u16 x,y;
	for(y=0; y<BG_SCREEN_CY; y++)
	{
		for(x=0; x<BG_SCREEN_CX; x++)
		{
			p->txt[y][x] = ' ';
		}
	}
	BGDrawCls(BG3.map, ' ');

	ScreenLoadSRAM();
	ScreenDrawAll();

	BGDrawCls(BG2.map, 0x00);
	ScreenDrawFrame();
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenDrawFrame()
{
	BGDrawStr(BG2.map, 0, 18, (u8*)"------------------------------");
	BGDrawStr(BG2.map, 0, 19, (u8*)" +-[].,><                     ");
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenEdit(u16 x, u16 y, u16 chr)
{
	Screen.txt[y][x] = (u8)chr;
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenDraw(u16 x, u16 y)
{
	BGDrawChr(BG3.map, x, y, BG3.col, (u16)Screen.txt[y][x]);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenDrawAll()
{
	u16 x,y;

	for(y=0; y<BG_SCREEN_CY; y++)
	{
		for(x=0; x<BG_SCREEN_CX; x++)
		{
			ScreenDraw(x, y);
		}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CursorInit()
{
	ST_CURSOR* p = &Cursor;

	p->x       =  0;
	p->y       = 19;
	p->cx      =  8;
	p->cy      =  8;
	p->minX    =  0;
	p->maxX    =  8;
	p->sprNo   =  1;
	p->sprTile =  1;
	p->sprSize = SP_SIZE_8;
	p->sprForm = SP_SQUARE;

	SpriteSetSize(p->sprNo, p->sprSize, p->sprForm, SP_COLOR_16);
	SpriteSetChr (p->sprNo, p->sprTile);
}
//---------------------------------------------------------------------------
EWRAM_CODE void WriterInit()
{
	ST_WRITER* p = &Writer;

	p->x       =  0;
	p->y       =  0;
	p->cx      =  8;
	p->cy      =  8;
	p->minX    =  0;
	p->maxX    = 29;
	p->minY    =  0;
	p->maxY    = 17;
	p->sprNo   =  2;
	p->sprTile =  2;
	p->sprSize = SP_SIZE_8;
	p->sprForm = SP_SQUARE;

	SpriteSetSize(p->sprNo, p->sprSize, p->sprForm, SP_COLOR_16);
	SpriteSetChr (p->sprNo, p->sprTile);
}
//---------------------------------------------------------------------------
EWRAM_CODE void CursorDraw()
{
	ST_CURSOR* p = &Cursor;

	s16 x  = p->x * p->cx;
	s16 y  = p->y * p->cy;
	u16 no = p->sprNo;

	SpriteMove(no, x, y);
}
//---------------------------------------------------------------------------
EWRAM_CODE void WriterDraw()
{
	ST_WRITER* p = &Writer;

	s16 x  = p->x * p->cx;
	s16 y  = p->y * p->cy;
	u16 no = p->sprNo;

	SpriteMove(no, x, y);
}
//---------------------------------------------------------------------------
EWRAM_CODE void StateInput()
{
	WaitForVsync();
	Key = KeyGet();

	if(Key == KEY_NONE)
	{
		return;
	}

	if(Key & KEY_START)
	{
		State = STATE_RUN;
		return;
	}

	if(Key & KEY_SELECT)
	{
//		ScreenSaveSRAM();
//		ScreenLoadSRAM();
		return;
	}

	if(Key & KEY_LR)
	{
		CursorInput();
	}

	WriterInput();
}
//---------------------------------------------------------------------------
EWRAM_CODE void CursorInput()
{
	ST_CURSOR* p = &Cursor;

	if(Key & KEY_L)
	{
		if(p->x <= p->minX)
		{
			return;
		}

		p->x--;
	}

	if(Key & KEY_R)
	{
		if(p->x >= p->maxX)
		{
			return;
		}

		p->x++;
	}

	CursorDraw();
}
//---------------------------------------------------------------------------
EWRAM_CODE void WriterInput()
{
	ST_CURSOR* c = &Cursor;
	ST_WRITER* w = &Writer;

	u16 chr = BGGetChr(BG2.map, c->x, c->y);

	if(Key & KEY_A)
	{
		ScreenEdit(w->x, w->y, chr);
		ScreenDraw(w->x, w->y);

		Key |= KEY_RIGHT;
	}

	if(Key & KEY_MOVE)
	{
		if( WriterIsMove() == TRUE )
		{
			WriterMove();
			WriterDraw();
		}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE bool WriterIsMove()
{
	ST_WRITER* p = &Writer;

	s16 tx1 = p->x;
	s16 ty1 = p->y;

	if(Key & KEY_UP   ) ty1--;
	if(Key & KEY_DOWN ) ty1++;
	if(Key & KEY_LEFT ) tx1--;
	if(Key & KEY_RIGHT) tx1++;

	if(tx1 < p->minX || ty1 < p->minY) return FALSE;
	if(tx1 > p->maxX || ty1 > p->maxY) return FALSE;

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void WriterMove()
{
	ST_WRITER* p = &Writer;

	if(Key & KEY_UP   ) p->y--;
	if(Key & KEY_DOWN ) p->y++;
	if(Key & KEY_LEFT ) p->x--;
	if(Key & KEY_RIGHT) p->x++;
}
//---------------------------------------------------------------------------
EWRAM_CODE void StateRun()
{
	BGDrawCls(BG3.map, ' ');

	BrainInit();
	BrainRun();
	KeyWait(KEY_A);

	ScreenDrawAll();
	State = STATE_INPUT;
}
//---------------------------------------------------------------------------
EWRAM_CODE void BrainInit()
{
	ST_BRAIN* p = &Brain;

	u16 i;
	for(i=0; i<BRAIN_MAX_MEM; i++)
	{
		p->mem[i] = 0x00;
	}

	p->point   = 0;
	p->deepCnt = 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void BrainRun()
{
	ST_BRAIN* b = &Brain;

	u8* pTxt  = (u8*)&Screen.txt[0][0];
	u8* epTxt = pTxt + (BG_SCREEN_CX * BG_SCREEN_CY);
	u8  curX, curY;
	
	curX = 0;
	curY = 0;

	while (pTxt < epTxt)
	{
//		TRACEOUT("%c\n", *pTxt);
		switch (*pTxt)
		{
		case ' ':
			//Ȃ
			break;

		case '>':
			b->point++;
			break;

		case '<':
			b->point--;
			break;

		case '+':
			b->mem[ b->point ]++;
			break;

		case '-':
			b->mem[ b->point ]--;
			break;

		case '.':
//			TRACEOUT("%c\n", b->mem[ b->point ]);
			BGDrawChr(BG3.map, curX++, curY, BG3.col, (u16)b->mem[ b->point ]);
			if(curX >= BG_SCREEN_CX)
			{
				curX = 0;
				curY++;

				if(curY >= BG_SCREEN_CY)
				{
					curY = 0;
				}
			}
			break;

		case ',':
			//TODO 
			break;

		case '[':
			if(b->mem[ b->point ] == 0x00)
			{
				pTxt++;
				for(b->deepCnt=1; b->deepCnt != 0; pTxt++)
				{
					if(*pTxt == '['){ b->deepCnt++; }
					if(*pTxt == ']'){ b->deepCnt--; }
				}
			}
			break;

		case ']':
			if (b->mem[ b->point ] != 0x00)
			{
				pTxt--;
				for(b->deepCnt=1; b->deepCnt != 0; pTxt--)
				{
					if(*pTxt == '['){ b->deepCnt--; }
					if(*pTxt == ']'){ b->deepCnt++; }
				}
			}
	        break;

		default:
			DebugMessage("Err: BrainRun");
			break;

		} //switch

		pTxt++;
	} //while
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenSaveSRAM()
{
	u8* p        = (u8*)&Screen;
	u32 size     = sizeof(ST_SCREEN);
	u32 writeCnt = 0;

	u32 i;
	for(i=0; i<size; i++)
	{
		SRAMWrite8(writeCnt++, *p++);
	}

	TRACEOUT("Save OK [size = %x]\n", writeCnt);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenLoadSRAM()
{
	u8* p        = (u8*)&Screen;
	u32 size     = sizeof(ST_SCREEN);
	u32 readCnt = 0;

	u32 i;
	for(i=0; i<size; i++)
	{
		*p++ = SRAMRead8(readCnt++);
	}

	TRACEOUT("Load OK [size = %x]\n", readCnt);
}
